// Copyright (c) 2014 Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of ACG (www.acg.cs.tau.ac.il).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Author(s): Michal Kleinbort <balasmic@post.tau.ac.il>
//            Oren Salzman <orenzalz@post.tau.ac.il >          

#ifndef BITSET_PAIRS_H__RTG
#define BITSET_PAIRS_H__RTG

#include <boost/dynamic_bitset.hpp>
#include "pairs.h"

namespace RTG
{

    /** \brief A class for storing the already reported pairs,
            which uses a bitset of the pairs.
	**/
    /*! The data structure is a bitset, where pair (i,j) is mapped to an index k.
    */
	class BitSetPairsStruct : public PairsStruct
    {
    public:
		typedef boost::dynamic_bitset<>           Pairs;
		
        /** \brief Constructor */
        BitSetPairsStruct(unsigned int n, bool add_once = true) 
			: PairsStruct(n, add_once),
			  pairs_((add_once)?((n/2) * (n - 1)):(n*(n-1)))
        {}

		/** \brief Check whether a pair exists in the structure */
		virtual bool pair_exists(int i, int j) const
		{
			assert (i != j);
            return (i < j) ? 
                    pairs_.test(get_index(i,j)) :
                    pairs_.test(get_index(j,i));;
		}
		
		/** \brief Add pair to the structure */
		virtual void add_pair(int i, int j)
		{
			assert (i != j);
            if (add_once_)
            {
                if (i < j) 
                    pairs_.set(get_index(i,j));
                else
                    pairs_.set(get_index(j,i));
            }
            else
            {
                pairs_.set(get_index(i,j));
                pairs_.set(get_index(j,i));
            }
			return;
		}

		/** \brief Clear the datastructure */
        virtual void clear(void)
		{
		    pairs_.clear();
		}

		/** \brief places the indices of the neighbors of index i in the given vector */
		virtual void neighbors_of_i(int i, std::vector<int>& neighborsVec) const
		{
            for (std::size_t j=0; j < num_of_points_; ++j)
				if (pairs_.test(get_index(i,j)))
					neighborsVec.push_back(j);
		}

		/** \brief places the reported pairs of indices in the given vector */
		virtual void all_pairs(std::vector<std::pair<int, int>>& pairsVec) const
		{
            for (std::size_t i(0); i < num_of_points_; ++i)
			{
                for (std::size_t j=i+1; j < num_of_points_; ++j)
					if (pairs_.test(get_index(i,j)))
						pairsVec.push_back(std::make_pair(i, j));
			}
		}
	 

	 private:
		
		std::size_t  get_index(int i, int j) const
		{
			int n = num_of_points_;
			if (add_once_)
				return (std::size_t)((2*n-i-1) * 0.5 * i + (j - i - 1));
			else //add_twice
				return ((n-1)*i + j);
		}

	 private:
		/// \brief A data structure holding the reported pairs
        Pairs pairs_;
		
    };

} //namespace RTG



#endif //BITSET_PAIRS_H__RTG